@@ -15,14 +15,16 @@
 -----------------------------------------------------------------------------
 
 module GF.Grammar.Grammar (
-        SourceGrammar, SourceModInfo(..), SourceModule, ModuleType(..),
-        emptySourceGrammar, mGrammar, modules, prependModule, moduleMap,
+        -- ** Grammar modules
+        Grammar, ModuleName, Module, ModuleInfo(..),
+        SourceGrammar, SourceModInfo, SourceModule,
+        ModuleType(..),
+        emptyGrammar, mGrammar, modules, prependModule, moduleMap,
 
         MInclude (..), OpenSpec(..),
         extends, isInherited, inheritAll,
-        openedModule, depPathModule, allDepsModule, partOfGrammar,
-        allExtends, allExtendsPlus,
-        searchPathModule,
+        openedModule, allDepsModule, partOfGrammar, depPathModule,
+        allExtends, allExtendsPlus, --searchPathModule,
         
         lookupModule,
         isModAbs, isModRes, isModCnc,
@@ -32,17 +34,16 @@ module GF.Grammar.Grammar (
         abstractOfConcrete,
 
         ModuleStatus(..),
-        
-        PMCFG(..), Production(..), FId, FunId, SeqId, LIndex, Sequence,
-        
+
+        -- ** Judgements
         Info(..),
-        Location(..), L(..), unLoc, noLoc,
+        -- ** Terms
+        Term(..),
         Type,
         Cat,
         Fun,
         QIdent,
         BindType(..),
-        Term(..),
         Patt(..),
         TInfo(..),
         Label(..),
@@ -58,75 +59,94 @@ module GF.Grammar.Grammar (
         Altern,
         Substitution,
         varLabel, tupleLabel, linLabel, theLinLabel,
-        ident2label, label2ident
+        ident2label, label2ident,
+        -- ** Source locations
+        Location(..), L(..), unLoc, noLoc, ppLocation, ppL,
+
+        -- ** PMCFG        
+        PMCFG(..), Production(..), FId, FunId, SeqId, LIndex, Sequence
         ) where
 
 import GF.Infra.Ident
 import GF.Infra.Option ---
+import GF.Infra.Location
 
 import GF.Data.Operations
 
 import PGF.Internal (FId, FunId, SeqId, LIndex, Sequence, BindType(..))
 
---import Data.List
-import Data.Array.IArray
-import Data.Array.Unboxed
+import Data.Array.IArray(Array)
+import Data.Array.Unboxed(UArray)
 import qualified Data.Map as Map
---import qualified Data.Set as Set
---import qualified Data.IntMap as IntMap
-import Text.PrettyPrint
---import System.FilePath
---import Control.Monad.Identity
+import qualified Data.Set as Set
+import GF.Text.Pretty
 
 
-
-data SourceGrammar = MGrammar { 
-    moduleMap :: Map.Map Ident SourceModInfo,
-    modules :: [SourceModule]
+-- | A grammar is a self-contained collection of grammar modules
+data Grammar = MGrammar { 
+    moduleMap :: Map.Map ModuleName ModuleInfo,
+    modules :: [Module]
   }
 
-data SourceModInfo = ModInfo {
+-- | Modules
+type Module = (ModuleName, ModuleInfo)
+
+data ModuleInfo = ModInfo {
     mtype   :: ModuleType,
     mstatus :: ModuleStatus,
     mflags  :: Options,
-    mextend :: [(Ident,MInclude)],
-    mwith   :: Maybe (Ident,MInclude,[(Ident,Ident)]),
+    mextend :: [(ModuleName,MInclude)],
+    mwith   :: Maybe (ModuleName,MInclude,[(ModuleName,ModuleName)]),
     mopens  :: [OpenSpec],
-    mexdeps :: [Ident],
+    mexdeps :: [ModuleName],
     msrc    :: FilePath,
     mseqs   :: Maybe (Array SeqId Sequence),
     jments  :: Map.Map Ident Info
   }
 
-type SourceModule = (Ident, SourceModInfo)
+type SourceGrammar = Grammar
+type SourceModule = Module
+type SourceModInfo = ModuleInfo
+
+instance HasSourcePath ModuleInfo where sourcePath = msrc
 
 -- | encoding the type of the module
 data ModuleType = 
     MTAbstract 
   | MTResource
-  | MTConcrete Ident
+  | MTConcrete ModuleName
   | MTInterface
-  | MTInstance (Ident,MInclude)
+  | MTInstance (ModuleName,MInclude)
   deriving (Eq,Show)
 
 data MInclude = MIAll | MIOnly [Ident] | MIExcept [Ident]
   deriving (Eq,Show)
 
-extends :: SourceModInfo -> [Ident]
+extends :: ModuleInfo -> [ModuleName]
 extends = map fst . mextend
 
 isInherited :: MInclude -> Ident -> Bool
-isInherited c i = case c of
-  MIAll -> True
-  MIOnly is -> elem i is
-  MIExcept is -> notElem i is
+isInherited c =
+  case c of
+    MIAll -> const True
+    MIOnly is -> elemOrd is
+    MIExcept is -> not . elemOrd is
 
-inheritAll :: Ident -> (Ident,MInclude)
+-- | Faster version of `elem`, using a `Set`.
+-- Make sure you give this the first argument _outside_ of the inner loop
+--
+-- Example:
+-- > myIntersection xs ys = filter (elemOrd xs) ys
+elemOrd :: Ord a => [a] -> a -> Bool
+elemOrd list = (`Set.member` set)
+  where set = Set.fromList list
+
+inheritAll :: ModuleName -> (ModuleName,MInclude)
 inheritAll i = (i,MIAll)
 
 data OpenSpec = 
-   OSimple Ident
- | OQualif Ident Ident
+   OSimple ModuleName
+ | OQualif ModuleName ModuleName
   deriving (Eq,Show)
 
 data ModuleStatus = 
@@ -134,13 +154,13 @@ data ModuleStatus =
  | MSIncomplete 
   deriving (Eq,Ord,Show)
 
-openedModule :: OpenSpec -> Ident
+openedModule :: OpenSpec -> ModuleName
 openedModule o = case o of
   OSimple m -> m
   OQualif _ m -> m
 
 -- | initial dependency list
-depPathModule :: SourceModInfo -> [OpenSpec]
+depPathModule :: ModuleInfo -> [OpenSpec]
 depPathModule m = fors m ++ exts m ++ mopens m
   where
     fors m = 
@@ -151,7 +171,7 @@ depPathModule m = fors m ++ exts m ++ mopens m
     exts m = map OSimple (extends m)
 
 -- | all dependencies
-allDepsModule :: SourceGrammar -> SourceModInfo -> [OpenSpec]
+allDepsModule :: Grammar -> ModuleInfo -> [OpenSpec]
 allDepsModule gr m = iterFix add os0 where
   os0 = depPathModule m
   add os = [m | o <- os, Just n <- [lookup (openedModule o) mods], 
@@ -159,21 +179,21 @@ allDepsModule gr m = iterFix add os0 where
   mods = modules gr
 
 -- | select just those modules that a given one depends on, including itself
-partOfGrammar :: SourceGrammar -> (Ident,SourceModInfo) -> SourceGrammar
+partOfGrammar :: Grammar -> Module -> Grammar
 partOfGrammar gr (i,m) = mGrammar [mo | mo@(j,_) <- mods, elem j modsFor]
   where
     mods = modules gr
     modsFor = (i:) $ map openedModule $ allDepsModule gr m
 
 -- | all modules that a module extends, directly or indirectly, with restricts
-allExtends :: SourceGrammar -> Ident -> [SourceModule]
+allExtends :: Grammar -> ModuleName -> [Module]
 allExtends gr m =
   case lookupModule gr m of
     Ok mi -> (m,mi) : concatMap (allExtends gr . fst) (mextend mi)
     _     -> []
 
 -- | the same as 'allExtends' plus that an instance extends its interface
-allExtendsPlus :: SourceGrammar -> Ident -> [Ident]
+allExtendsPlus :: Grammar -> ModuleName -> [ModuleName]
 allExtendsPlus gr i =
   case lookupModule gr i of
     Ok m -> i : concatMap (allExtendsPlus gr) (exts m)
@@ -181,39 +201,40 @@ allExtendsPlus gr i =
   where
     exts m = extends m ++ [j | MTInstance (j,_) <- [mtype m]]
 
--- | initial search path: the nonqualified dependencies
-searchPathModule :: SourceModInfo -> [Ident]
-searchPathModule m = [i | OSimple i <- depPathModule m]
+-- -- | initial search path: the nonqualified dependencies
+-- searchPathModule :: ModuleInfo -> [ModuleName]
+-- searchPathModule m = [i | OSimple i <- depPathModule m]
 
+prependModule :: Grammar -> Module -> Grammar
 prependModule (MGrammar mm ms) im@(i,m) = MGrammar (Map.insert i m mm) (im:ms)
 
-emptySourceGrammar :: SourceGrammar
-emptySourceGrammar = mGrammar []
+emptyGrammar = mGrammar []
 
+mGrammar :: [Module] -> Grammar
 mGrammar ms = MGrammar (Map.fromList ms) ms
 
 
 -- | we store the module type with the identifier
 
-abstractOfConcrete :: ErrorMonad m => SourceGrammar -> Ident -> m Ident
+abstractOfConcrete :: ErrorMonad m => Grammar -> ModuleName -> m ModuleName
 abstractOfConcrete gr c = do
   n <- lookupModule gr c
   case mtype n of
     MTConcrete a -> return a
-    _ -> raise $ render (text "expected concrete" <+> ppIdent c)
+    _ -> raise $ render ("expected concrete" <+> c)
 
-lookupModule :: ErrorMonad m => SourceGrammar -> Ident -> m SourceModInfo
+lookupModule :: ErrorMonad m => Grammar -> ModuleName -> m ModuleInfo
 lookupModule gr m = case Map.lookup m (moduleMap gr) of
   Just i  -> return i
-  Nothing -> raise $ render (text "unknown module" <+> ppIdent m <+> text "among" <+> hsep (map (ppIdent . fst) (modules gr)))
+  Nothing -> raise $ render ("unknown module" <+> m <+> "among" <+> hsep (map fst (modules gr)))
 
-isModAbs :: SourceModInfo -> Bool
+isModAbs :: ModuleInfo -> Bool
 isModAbs m =
   case mtype m of
     MTAbstract -> True
     _          -> False
 
-isModRes :: SourceModInfo -> Bool
+isModRes :: ModuleInfo -> Bool
 isModRes m =
   case mtype m of
     MTResource   -> True
@@ -221,7 +242,7 @@ isModRes m =
     MTInstance _ -> True
     _            -> False
 
-isModCnc :: SourceModInfo -> Bool
+isModCnc :: ModuleInfo -> Bool
 isModCnc m =
   case mtype m of
     MTConcrete _ -> True
@@ -247,49 +268,49 @@ sameMType m n =
     _                            -> m == n
 
 -- | don't generate code for interfaces and for incomplete modules
-isCompilableModule :: SourceModInfo -> Bool
+isCompilableModule :: ModuleInfo -> Bool
 isCompilableModule m =
   case mtype m of
     MTInterface -> False
     _           -> mstatus m == MSComplete
 
 -- | interface and "incomplete M" are not complete
-isCompleteModule :: SourceModInfo -> Bool
+isCompleteModule :: ModuleInfo -> Bool
 isCompleteModule m = mstatus m == MSComplete && mtype m /= MTInterface
 
 
 -- | all abstract modules sorted from least to most dependent
-allAbstracts :: SourceGrammar -> [Ident]
+allAbstracts :: Grammar -> [ModuleName]
 allAbstracts gr = 
  case topoTest [(i,extends m) | (i,m) <- modules gr, mtype m == MTAbstract] of
    Left  is     -> is
-   Right cycles -> error $ render (text "Cyclic abstract modules:" <+> vcat (map (hsep . map ppIdent) cycles))
+   Right cycles -> error $ render ("Cyclic abstract modules:" <+> vcat (map hsep cycles))
 
 -- | the last abstract in dependency order (head of list)
-greatestAbstract :: SourceGrammar -> Maybe Ident
+greatestAbstract :: Grammar -> Maybe ModuleName
 greatestAbstract gr =
   case allAbstracts gr of
     [] -> Nothing
     as -> return $ last as
 
 -- | all resource modules
-allResources :: SourceGrammar -> [Ident]
+allResources :: Grammar -> [ModuleName]
 allResources gr = [i | (i,m) <- modules gr, isModRes m || isModCnc m]
 
 -- | the greatest resource in dependency order
-greatestResource :: SourceGrammar -> Maybe Ident
+greatestResource :: Grammar -> Maybe ModuleName
 greatestResource gr =
   case allResources gr of
     [] -> Nothing
-    a  -> return $ head a ---- why not last as in Abstract? works though AR 24/5/2008
+    mo:_ -> Just mo ---- why not last as in Abstract? works though AR 24/5/2008
 
 -- | all concretes for a given abstract
-allConcretes :: SourceGrammar -> Ident -> [Ident]
+allConcretes :: Grammar -> ModuleName -> [ModuleName]
 allConcretes gr a =
   [i | (i, m) <- modules gr, mtype m == MTConcrete a, isCompleteModule m]
 
 -- | all concrete modules for any abstract
-allConcreteModules :: SourceGrammar -> [Ident]
+allConcreteModules :: Grammar -> [ModuleName]
 allConcreteModules gr =
   [i | (i, m) <- modules gr, MTConcrete _ <- [mtype m], isCompleteModule m]
 
@@ -322,38 +343,21 @@ data Info =
  | ResValue (L Type)                             -- ^ (/RES/) to mark parameter constructors for lookup
  | ResOper  (Maybe (L Type)) (Maybe (L Term))    -- ^ (/RES/)
 
- | ResOverload [Ident] [(L Type,L Term)]         -- ^ (/RES/) idents: modules inherited
+ | ResOverload [ModuleName] [(L Type,L Term)]         -- ^ (/RES/) idents: modules inherited
 
 -- judgements in concrete syntax
  | CncCat  (Maybe (L Type))             (Maybe (L Term)) (Maybe (L Term)) (Maybe (L Term)) (Maybe PMCFG) -- ^ (/CNC/) lindef ini'zed, 
  | CncFun  (Maybe (Ident,Context,Type)) (Maybe (L Term))                  (Maybe (L Term)) (Maybe PMCFG) -- ^ (/CNC/) type info added at 'TC'
 
 -- indirection to module Ident
- | AnyInd Bool Ident                         -- ^ (/INDIR/) the 'Bool' says if canonical
+ | AnyInd Bool ModuleName                        -- ^ (/INDIR/) the 'Bool' says if canonical
   deriving Show
 
-data Location 
-  = NoLoc
-  | Local Int Int
-  | External FilePath Location
-  deriving (Show,Eq,Ord)
-
-data L a = L Location a  -- location information
-  deriving Show
-
-instance Functor L where
-  fmap f (L loc x) = L loc (f x)
-
-unLoc :: L a -> a
-unLoc (L _ x) = x
-
-noLoc = L NoLoc
-
 type Type = Term
 type Cat  = QIdent
 type Fun  = QIdent
 
-type QIdent = (Ident,Ident)
+type QIdent = (ModuleName,Ident)
 
 data Term =
    Vr Ident                      -- ^ variable
@@ -462,7 +466,7 @@ type Equation = ([Patt],Term)
 type Labelling = (Label, Type) 
 type Assign = (Label, (Maybe Type, Term)) 
 type Case = (Patt, Term) 
-type Cases = ([Patt], Term) 
+--type Cases = ([Patt], Term) 
 type LocalDef = (Ident, (Maybe Type, Term))
 
 type Param = (Ident, Context) 
