@@ -4,8 +4,8 @@ module PGFService(cgiMain,cgiMain',getPath,
                   Caches,pgfCache,newPGFCache,flushPGFCache,listPGFCache) where
 
 import PGF (PGF,Labels,CncLabels)
+import GF.Text.Lexing
 import qualified PGF
-import PGF.Lexing
 import Cache
 import CGIUtils(outputJSONP,outputPlain,outputHTML,outputText,
                 outputBinary,outputBinary',
@@ -39,7 +39,7 @@ import Control.Monad.State(State,evalState,get,put)
 import Control.Monad.Catch(bracket_)
 import Data.Char
 --import Data.Function (on)
-import Data.List ({-sortBy,-}intersperse,mapAccumL,nub,isSuffixOf,nubBy)
+import Data.List ({-sortBy,-}intersperse,mapAccumL,nub,isSuffixOf,nubBy,stripPrefix)
 import qualified Data.Map as Map
 import Data.Maybe
 import System.Random
@@ -151,24 +151,37 @@ getFile get path =
 cpgfMain qsem command (t,(pgf,pc)) =
   case command of
     "c-parse"       -> withQSem qsem $
-                       out t=<< join (parse # input % start % limit % treeopts)
+                       out t=<< join (parse # input % cat % start % limit % treeopts)
+    "c-parseToChart"-> withQSem qsem $
+                       out t=<< join (parseToChart # input % cat % limit)
     "c-linearize"   -> out t=<< lin # tree % to
+    "c-bracketedLinearize"
+                    -> out t=<< bracketedLin # tree % to
     "c-linearizeAll"-> out t=<< linAll # tree % to
     "c-translate"   -> withQSem qsem $
-                       out t=<<join(trans # input % to % start % limit%treeopts)
+                       out t=<<join(trans # input % cat % to % start % limit % treeopts)
     "c-lookupmorpho"-> out t=<< morpho # from1 % textInput
+    "c-lookupcohorts"->out t=<< cohorts # from1 % getInput "filter" % textInput
     "c-flush"       -> out t=<< flush
     "c-grammar"     -> out t grammar
-    "c-abstrtree"   -> outputGraphviz=<< C.graphvizAbstractTree pgf # tree
-    "c-parsetree"   -> outputGraphviz=<< C.graphvizParseTree . snd # from1 %tree
-    "c-wordforword" -> out t =<< wordforword # input % to
+    "c-abstrtree"   -> outputGraphviz=<< C.graphvizAbstractTree pgf C.graphvizDefaults # tree
+    "c-parsetree"   -> outputGraphviz=<< (\cnc -> C.graphvizParseTree cnc C.graphvizDefaults) . snd # from1 % tree
+    "c-wordforword" -> out t =<< wordforword # input % cat % to
     _               -> badRequest "Unknown command" command
   where
     flush = liftIO $ do --modifyMVar_ pc $ const $ return Map.empty
                         performGC
                         return $ showJSON ()
 
-    cat = C.startCat pgf
+    cat :: CGI C.Type
+    cat =
+       do mcat  <- getInput1 "cat"
+          case mcat of
+            Nothing -> return (C.startCat pgf)
+            Just cat -> case C.readType cat of
+                          Nothing  -> badRequest "Bad category" cat
+                          Just typ -> return typ
+
     langs = C.languages pgf
 
     grammar = showJSON $ makeObj
@@ -179,8 +192,8 @@ cpgfMain qsem command (t,(pgf,pc)) =
       where
         languages = [makeObj ["name".= l] | (l,_)<-Map.toList langs]
 
-    parse input@((from,_),_) start mlimit (trie,json) =
-        do r <- parse' start mlimit input
+    parse input@((from,_),_) cat start mlimit (trie,json) =
+        do r <- parse' cat start mlimit input
            return $ showJSON [makeObj ("from".=from:jsonParseResult json r)]
 
     jsonParseResult json = either bad good
@@ -190,11 +203,12 @@ cpgfMain qsem command (t,(pgf,pc)) =
         tp (tree,prob) = makeObj (addTree json tree++["prob".=prob])
 
     -- Without caching parse results:
-    parse' start mlimit ((from,concr),input) =
-        return $ maybe id take mlimit . drop start # cparse
+    parse' cat start mlimit ((from,concr),input) =
+        case C.parseWithHeuristics concr cat input (-1) callbacks of
+          C.ParseOk ts        -> return (Right (maybe id take mlimit (drop start ts)))
+          C.ParseFailed _ tok -> return (Left tok)
+          C.ParseIncomplete   -> return (Left "")
       where
-      --cparse = C.parse concr cat input
-        cparse = C.parseWithHeuristics concr cat input (-1) callbacks
         callbacks = maybe [] cb $ lookup (C.abstractName pgf) C.literalCallbacks
         cb fs = [(cat,f pgf (from,concr) input)|(cat,f)<-fs]
 {-
@@ -214,6 +228,35 @@ cpgfMain qsem command (t,(pgf,pc)) =
             purge r@(_,t') = if diffUTCTime t t'<120 then Just r else Nothing
                              -- remove unused parse results after 2 minutes
 -}
+
+    parseToChart ((from,concr),input) cat mlimit =
+      do r <- case C.parseToChart concr cat input (-1) callbacks (fromMaybe 5 mlimit) of
+                C.ParseOk chart     -> return (good chart)
+                C.ParseFailed _ tok -> return (bad tok)
+                C.ParseIncomplete   -> return (bad "")
+         return $ showJSON [makeObj ("from".=from:r)]
+      where
+        callbacks = maybe [] cb $ lookup (C.abstractName pgf) C.literalCallbacks
+        cb fs = [(cat,f pgf (from,concr) input)|(cat,f)<-fs]
+
+        bad  err           = ["parseFailed".=err]
+        good (roots,chart) = ["roots".=showJSON roots,
+                              "chart".=makeObj [show fid .= mkChartObj inf | (fid,inf)<-Map.toList chart]]
+
+        mkChartObj (brackets,prods,cat) =
+          makeObj ["brackets".=map mkChartBracket brackets
+                  ,"prods"   .=map mkChartProd prods
+                  ,"cat"     .=cat
+                  ]
+
+        mkChartBracket (s,e,ann) =
+          makeObj ["start".=s,"end".=e,"ann".=ann]
+
+        mkChartProd (expr,args,prob) =
+          makeObj ["tree".=expr,"args".=map mkChartPArg args,"prob".=prob]
+
+        mkChartPArg (C.PArg _ fid) = showJSON fid
+
     linAll tree to = showJSON (linAll' tree to)
     linAll' tree (tos,unlex) =
         [makeObj ["to".=to,
@@ -223,8 +266,12 @@ cpgfMain qsem command (t,(pgf,pc)) =
     lin' tree (tos,unlex) =
         [makeObj ["to".=to,"text".=unlex (C.linearize c tree)]|(to,c)<-tos]
 
-    trans input@((from,_),_) to start mlimit (trie,jsontree) =
-      do parses <- parse' start mlimit input
+    bracketedLin tree to = showJSON (bracketedLin' tree to)
+    bracketedLin' tree (tos,unlex) =
+        [makeObj ["to".=to,"brackets".=showJSON (C.bracketedLinearize c tree)]|(to,c)<-tos]
+
+    trans input@((from,_),_) cat to start mlimit (trie,jsontree) =
+      do parses <- parse' cat start mlimit input
          return $
            showJSON [ makeObj ["from".=from,
                                "translations".= jsonParses parses]]
@@ -238,11 +285,27 @@ cpgfMain qsem command (t,(pgf,pc)) =
                                     | (tree,prob) <- parses]
 
     morpho (from,concr) input =
-        showJSON [makeObj ["lemma".=l,"analysis".=a,"prob".=p]|(l,a,p)<-ms]
-      where ms = C.lookupMorpho concr input
+        showJSON [makeObj ["lemma".=l
+                          ,"analysis".=a
+                          ,"prob".=p]
+                     | (l,a,p)<-C.lookupMorpho concr input]
 
+    cohorts (from,concr) filter input =
+      showJSON [makeObj ["start" .=showJSON s
+                        ,"word"  .=showJSON w
+                        ,"morpho".=showJSON [makeObj ["lemma".=l
+                                                     ,"analysis".=a
+                                                     ,"prob".=p] 
+                                                | (l,a,p)<-ms]
+                        ,"end"   .=showJSON e
+                        ]
+                   | (s,w,ms,e) <- (case filter of
+                                      Just "longest" -> C.filterLongest
+                                      Just "best"    -> C.filterBest
+                                      _              -> id)
+                                     (C.lookupCohorts concr input)]
 
-    wordforword input@((from,_),_) = jsonWFW from . wordforword' input
+    wordforword input@((from,_),_) cat = jsonWFW from . wordforword' input cat
 
     jsonWFW from rs =
       showJSON
@@ -252,7 +315,7 @@ cpgfMain qsem command (t,(pgf,pc)) =
                                         [makeObj["to".=to,"text".=text]
                                          | (to,text)<-rs]]]]]
 
-    wordforword' inp@((from,concr),input) (tos,unlex) =
+    wordforword' inp@((from,concr),input) cat (tos,unlex) =
         [(to,unlex . unwords $ map (lin_word' c) pws)
          |let pws=map parse_word' (words input),(to,c)<-tos]
       where
@@ -272,10 +335,14 @@ cpgfMain qsem command (t,(pgf,pc)) =
             maybe (Left ("["++w++"]")) Right $
             msum [parse1 w,parse1 ow,morph w,morph ow]
           where
-            ow = if w==lw then capitInit w else lw
-            lw = uncapitInit w
-            parse1 = either (const Nothing) (fmap fst . listToMaybe) .
-                     C.parse concr cat
+            ow = case w of
+                   c:cs | isLower c -> toUpper c : cs
+                        | isUpper c -> toLower c : cs
+                   s                -> s
+
+            parse1 s = case C.parse concr cat s of
+                         C.ParseOk ((t,_):ts) -> Just t
+                         _                    -> Nothing
             morph w = listToMaybe
                         [t | (f,a,p)<-C.lookupMorpho concr w,
                              t<-maybeToList (C.readExpr f)]
@@ -293,7 +360,7 @@ cpgfMain qsem command (t,(pgf,pc)) =
     from1 = maybe (missing "from") return =<< from'
     from' = getLang "from"
 
-    to = (,) # getLangs "to" % unlexerC
+    to = (,) # getLangs "to" % unlexerC (const False)
 
     getLangs = getLangs' readLang
     getLang = getLang' readLang
@@ -308,8 +375,7 @@ cpgfMain qsem command (t,(pgf,pc)) =
               let t = C.readExpr s
               maybe (badRequest "bad tree" s) return t
 
-    --c_lexer concr = lexer
-    c_lexer concr = ilexer (not . null . C.lookupMorpho concr)
+    c_lexer concr = lexer (not . null . C.lookupMorpho concr)
 
 --------------------------------------------------------------------------------
 
@@ -338,62 +404,36 @@ instance ToATree C.Expr where
 --------------------------------------------------------------------------------
 -- * Lexing
 
--- | Lexers with a text lexer that tries to be a more clever with the first word
-ilexer good = lexer' uncap
-  where
-    uncap s = case span isUpper s of
-                ([c],r) | not (good s) -> toLower c:r
-                _ -> s
-
 -- | Standard lexers
-lexer = lexer' uncapitInit
-
-lexer' uncap = maybe (return id) lexerfun =<< getInput "lexer" 
+lexer good = maybe (return id) lexerfun =<< getInput "lexer" 
   where
     lexerfun name =
-       case name of
-         "text"  -> return (unwords . lexText' uncap)
-         "code"  -> return (unwords . lexCode)
-         "mixed" -> return (unwords . lexMixed)
-         _ -> badRequest "Unknown lexer" name
+      case stringOp good ("lex"++name) of
+        Just fn -> return fn
+        Nothing -> badRequest "Unknown lexer" name
 
 
 type Unlexer = String->String
 
 -- | Unlexing for the C runtime system, &+ is already applied
-unlexerC :: CGI Unlexer
-unlexerC = maybe (return id) unlexerfun =<< getInput "unlexer"
-  where
-    unlexerfun name =
-       case name of
-         "text"  -> return (unlexText' . words)
-         "code"  -> return (unlexCode . words)
-         "mixed" -> return (unlexMixed . words)
-         "none"  -> return id
-         "id"    -> return id
-         _ -> badRequest "Unknown lexer" name
+unlexerC :: (String -> Bool) -> CGI Unlexer
+unlexerC = unlexer' id
 
--- | Unlex text, skipping the quality marker used by the App grammar
-unlexText' ("+":ws) = "+ "++unlexText ws
-unlexText' ("*":ws) = "* "++unlexText ws
-unlexText' ws       = unlexText ws
-
--- | Unlexing for the Haskell run-time, applying the &+ operator first
+-- | Unlexing for the Haskell runtime system, the default is to just apply &+
 unlexerH :: CGI Unlexer
-unlexerH = maybe (return doBind) unlexerfun =<< getInput "unlexer"
+unlexerH = unlexer' (unwords . bindTok . words) (const False)
+
+unlexer' defaultUnlexer good =
+    maybe (return defaultUnlexer) unlexerfun =<< getInput "unlexer"
   where
     unlexerfun name =
-       case name of
-         "text"  -> return (unlexText' . bind)
-         "code"  -> return (unlexCode . bind)
-         "mixed" -> return (unlexMixed . bind)
-         "none"  -> return id
-         "id"    -> return id
-         "bind"  -> return doBind
-         _ -> badRequest "Unknown lexer" name
-
-    doBind = unwords . bind
-    bind = bindTok . words
+       case stringOp good ("unlex"++name) of
+         Just fn -> return (fn . cleanMarker)
+         Nothing -> badRequest "Unknown unlexer" name
+    
+    cleanMarker ('+':cs) = cs
+    cleanMarker ('*':cs) = cs
+    cleanMarker cs       = cs
 
 --------------------------------------------------------------------------------
 -- * Haskell run-time functionality
@@ -408,7 +448,7 @@ pgfMain lcs@(alc,clc) path command tpgf@(t,pgf) =
       "linearizeTable" -> o =<< doLinearizeTabular pgf # tree % to
       "random"         -> o =<< join (doRandom pgf # cat % depth % limit % to)
       "generate"       -> o =<< doGenerate pgf # cat % depth % limit % to
-      "translate"      -> o =<< doTranslate pgf # input % cat %to%limit%treeopts
+      "translate"      -> o =<< doTranslate pgf # input % cat % to % limit % treeopts
       "translategroup" -> o =<< doTranslateGroup pgf # input % cat % to % limit
       "lookupmorpho"   -> o =<< doLookupMorpho pgf # from1 % textInput
       "grammar"        -> join $ doGrammar tpgf
@@ -431,8 +471,8 @@ pgfMain lcs@(alc,clc) path command tpgf@(t,pgf) =
                inp <- textInput
                return (fr,lex inp)
 
-    mlexer Nothing = lexer
-    mlexer (Just lang) = ilexer (PGF.isInMorpho morpho)
+    mlexer Nothing     = lexer (const False)
+    mlexer (Just lang) = lexer (PGF.isInMorpho morpho)
       where morpho = PGF.buildMorpho pgf lang
 
     tree :: CGI PGF.Tree
@@ -531,6 +571,8 @@ limit, depth :: CGI (Maybe Int)
 limit = readInput "limit"
 depth = readInput "depth"
 
+default_depth_server = 4
+
 start :: CGI Int
 start = maybe 0 id # readInput "start"
 
@@ -692,19 +734,16 @@ doComplete pgf (mfrom,input) mcat mlimit full = showJSON
     froms = maybe (PGF.languages pgf) (:[]) mfrom
     cat = fromMaybe (PGF.startCat pgf) mcat
 
-completionInfo :: PGF -> PGF.Token -> PGF.ParseState -> JSValue
-completionInfo pgf token pstate =
+completionInfo :: PGF -> PGF.Token -> [PGF.CId] -> JSValue
+completionInfo pgf token funs =
   makeObj
   ["token".= token
-  ,"funs" .= (map mkFun (nubBy ignoreFunIds funs))
+  ,"funs" .= map mkFun (nub funs)
   ]
   where
-    contInfo = PGF.getContinuationInfo pstate
-    funs = snd . head $ Map.toList contInfo -- always get [([],_)] ; funs :: [(fid,cid,seq)]
-    ignoreFunIds (_,cid1,seq1) (_,cid2,seq2) = (cid1,seq1) == (cid2,seq2)
-    mkFun (funid,cid,seq) = case PGF.functionType pgf cid of
+    mkFun cid = case PGF.functionType pgf cid of
       Just typ ->
-        makeObj [ {-"fid".=funid,-} "fun".=cid, "hyps".=hyps', "cat".=cat, "seq".=seq ]
+        makeObj [ {-"fid".=funid,-} "fun".=cid, "hyps".=hyps', "cat".=cat ]
         where
           (hyps,cat,_es) = PGF.unType typ
           hyps' = [ PGF.showType [] typ | (_,_,typ) <- hyps ]
@@ -744,7 +783,7 @@ doRandom pgf mcat mdepth mlimit to =
              | tree <- limit trees]
   where cat = fromMaybe (PGF.startCat pgf) mcat
         limit = take (fromMaybe 1 mlimit)
-        depth = fromMaybe 4 mdepth
+        depth = fromMaybe default_depth_server mdepth
 
 doGenerate :: PGF -> Maybe PGF.Type -> Maybe Int -> Maybe Int -> To -> JSValue
 doGenerate pgf mcat mdepth mlimit tos =
@@ -757,7 +796,7 @@ doGenerate pgf mcat mdepth mlimit tos =
     trees = PGF.generateAllDepth pgf cat (Just depth)
     cat = fromMaybe (PGF.startCat pgf) mcat
     limit = take (fromMaybe 1 mlimit)
-    depth = fromMaybe 4 mdepth
+    depth = fromMaybe default_depth_server mdepth
 
 doGrammar :: (UTCTime,PGF) -> Either IOError (UTCTime,l) -> Maybe (Accept Language) -> CGI CGIResult
 doGrammar (t1,pgf) elbls macc = out t $ showJSON $ makeObj
@@ -995,6 +1034,7 @@ instance JSON PGF.Trie where
     showJSON (PGF.Ap f [[]]) = makeObj ["fun".=f] -- leaf
 --  showJSON (PGF.Ap f [es]) = makeObj ["fun".=f,"children".=es] -- one alternative
     showJSON (PGF.Ap f alts) = makeObj ["fun".=f,"alts".=alts]
+    readJSON = error "PGF.Trie.readJSON intentionally not defined"
 
 instance JSON PGF.CId where
     readJSON x = readJSON x >>= maybe (fail "Bad language.") return . PGF.readLanguage
@@ -1006,10 +1046,19 @@ instance JSON PGF.Expr where
 
 instance JSON PGF.BracketedString where
     readJSON x = return (PGF.Leaf "")
-    showJSON (PGF.Bracket cat fid index fun _ bs) =
+    showJSON (PGF.Bracket cat fid _ index fun _ bs) =
         makeObj ["cat".=cat, "fid".=fid, "index".=index, "fun".=fun, "children".=bs]
     showJSON (PGF.Leaf s) = makeObj ["token".=s]
 
+#if C_RUNTIME
+instance JSON C.BracketedString where
+    readJSON x = return (C.Leaf "")
+    showJSON (C.Bracket cat fid index fun bs) =
+        makeObj ["cat".=cat, "fid".=fid, "index".=index, "fun".=fun, "children".=bs]
+    showJSON C.BIND     = makeObj ["bind".=True]
+    showJSON (C.Leaf s) = makeObj ["token".=s]
+#endif
+
 -- * PGF utilities
 {-
 cat :: PGF -> Maybe PGF.Type -> PGF.Type
@@ -1022,28 +1071,17 @@ parse' pgf input mcat mfrom =
         cat = fromMaybe (PGF.startCat pgf) mcat
 
 complete' :: PGF -> PGF.Language -> PGF.Type -> Maybe Int -> String
-         -> (PGF.BracketedString, String, Map.Map PGF.Token PGF.ParseState)
+         -> (PGF.BracketedString, String, Map.Map PGF.Token [PGF.CId])
 complete' pgf from typ mlimit input =
   let (ws,prefix) = tokensAndPrefix input
-      ps0 = PGF.initState pgf from typ
-      (ps,ws') = loop ps0 ws
-      bs       = snd (PGF.getParseOutput ps typ Nothing)
-  in if not (null ws')
-       then (bs, unwords (if null prefix then ws' else ws'++[prefix]), Map.empty)
-       else (bs, prefix, PGF.getCompletions ps prefix)
+  in PGF.complete pgf from typ (unwords ws) prefix
   where
-  --order = sortBy (compare `on` map toLower)
-
     tokensAndPrefix :: String -> ([String],String)
     tokensAndPrefix s | not (null s) && isSpace (last s) = (ws, "")
                       | null ws = ([],"")
                       | otherwise = (init ws, last ws)
         where ws = words s
 
-    loop ps []     = (ps,[])
-    loop ps (w:ws) = case PGF.nextState ps (PGF.simpleParseInput w) of
-                       Left  es -> (ps,w:ws)
-                       Right ps -> loop ps ws
 
 transfer lang = if "LaTeX" `isSuffixOf` show lang
                 then fold -- OpenMath LaTeX transfer
@@ -1056,7 +1094,7 @@ linearizeTabular pgf (tos,unlex) tree =
     [(to,lintab to (transfer to tree)) | to <- langs]
   where
     langs = if null tos then PGF.languages pgf else tos
-    lintab to t = [(p,map unlex (nub [t|(p',t)<-vs,p'==p]))|p<-ps]
+    lintab to t = [(p,map unlex (nub [t | (p',t)<-vs,p'==p])) | p<-ps]
       where
         ps = nub (map fst vs)
         vs = concat (PGF.tabularLinearizes pgf to t)
@@ -1069,16 +1107,23 @@ linearizeAndUnlex pgf (mto,unlex) tree =
     langs = if null mto then PGF.languages pgf else mto
 
 selectLanguage :: PGF -> Maybe (Accept Language) -> PGF.Language
-selectLanguage pgf macc = case acceptable of
-                            []  -> case PGF.languages pgf of
-                                     []  -> error "No concrete syntaxes in PGF grammar."
-                                     l:_ -> l
-                            Language c:_ -> fromJust (langCodeLanguage pgf c)
+selectLanguage pgf macc =
+    case acceptable of
+      []  -> case PGF.languages pgf of
+               []  -> error "No concrete syntaxes in PGF grammar."
+               ls@(l1:_) -> case [l | l<-ls, langPart pgf l==Just "Eng"] of
+                              eng:_ -> eng
+                              _ -> l1
+      Language c:_ -> fromJust (langCodeLanguage pgf c)
   where langCodes = mapMaybe (PGF.languageCode pgf) (PGF.languages pgf)
         acceptable = negotiate (map Language langCodes) macc
 
 langCodeLanguage :: PGF -> String -> Maybe PGF.Language
-langCodeLanguage pgf code = listToMaybe [l | l <- PGF.languages pgf, PGF.languageCode pgf l == Just code]
+langCodeLanguage pgf code =
+  listToMaybe [l | l <- PGF.languages pgf, PGF.languageCode pgf l == Just code]
+
+langPart pgf lang =
+  stripPrefix (PGF.showCId (PGF.abstractName pgf)) (PGF.showCId lang)
 
 -- * General utilities
 
