@@ -1,5 +1,3 @@
-
-
 var editor=element("editor");
 var compiler_output=element("compiler_output")
 
@@ -126,35 +124,70 @@ function draw_grammar_list() {
 	function rmpublic(file) {
 	    return function() { remove_public(file,draw_grammar_list) }
 	}
-	publiclist.appendChild(wrap("h3",text("Public grammars")))
-	if(files.length>0) {
-	    var unique_id=local.get("unique_id","-")
-	    var t=empty_class("table","grammar_list")
-	    for(var i in files) {
-		var file=files[i].path
-		var parts=file.split(/[-.]/)
-		var basename=parts[0]
-		var unique_name=parts[1]+"-"+parts[2]
-		var mine = my_grammar(unique_name)!=null
-		var del = mine
-		    ? delete_button(rmpublic(file),"Don't publish this grammar")
-		    : []
-		var tip = mine
-		    ? "This is a copy of your grammar"
-		    : "Click to download a copy of this grammar"
-		var modt=new Date(files[i].time)
-		var fmtmodt=modt.toDateString()+", "+modt.toTimeString().split(" ")[0]
-		var when=wrap_class("small","modtime",text(" "+fmtmodt))
-		t.appendChild(edtr([td(del),
-				    td(title(tip,
-					     a(jsurl('open_public("'+file+'")'),
-					       [text(basename)]))),
-				    td(when)]))
-	    }
-	    publiclist.appendChild(t)
+	var h=wrap("h3",text("Public grammars"))
+	var ordermenu=wrap("select",[option("Newest first","byAge"),
+				     option("Alphabetical","byName")])
+	ordermenu.value=local.get("publicOrder","byAge")
+	ordermenu.onchange=function(){
+	    local.put("publicOrder",ordermenu.value)
+	    if(n>1) show_grammars()
 	}
+	var n=files.length
+	var count=n==1 ? " (One grammar)" : " ("+n + " grammars)"
+	var t=table(tr([td(h),td(text(count)),td(ordermenu)]))
+	publiclist.appendChild(t)
+	for(var i in files) {
+	    var file=files[i]
+	    file.t=new Date(file.time)
+	    file.s=file.t.getTime()
+	}
+	function sort_grammars() {
+	    switch(ordermenu.value) {
+	    case "byAge":
+		files.sort((f1,f2)=>f2.s-f1.s)
+		break;
+	    case "byName":
+		files.sort((f1,f2)=>(f1.path>f2.path)-(f1.path<f2.path))
+	    }
+	}
+	var gt=empty_class("table","grammar_list")
+	publiclist.appendChild(gt)
+	function show_grammars() {
+	    clear(gt)
+	    if(files.length>0) {
+		sort_grammars()
+		var unique_id=local.get("unique_id","-")
+		for(var i in files) {
+		    var file=files[i].path
+		    var parts=file.split(/[-.]/)
+		    var basename=parts[0]
+		    var unique_name=parts[1]+"-"+parts[2]
+		    var mine = my_grammar(unique_name)!=null
+		    var from_me = parts[1] == unique_id
+		    var del = from_me || mine
+			? delete_button(rmpublic(file),"Remove this public grammar")
+			: []
+		    var tip = mine
+			? "This is a copy of your grammar"
+			: "Click to download a copy of this grammar"
+		    var modt=new Date(files[i].time)
+		    var fmtmodt=modt.toDateString()+", "+modt.toTimeString().split(" ")[0]
+		    var when=wrap_class("small","modtime",text(" "+fmtmodt))
+		    gt.appendChild(edtr([td(del),
+					 td(title(tip,
+						  a(jsurl('open_public("'+file+'")'),
+						    [text(basename)]))),
+					 td(text(files[i].comment||"")),
+					 td(when)]))
+		}
+	    }
 	else
 	    publiclist.appendChild(p(text("No public grammars are available.")))
+	    // This is outside the table so it won't be cleared,
+	    // but show_grammars is only called once then there is less
+	    // than 2 grammars, so it's OK.
+	}
+	show_grammars()
     }
     if(navigator.onLine)
 	gfcloud_public_json("ls-l",{},show_public,no_public)
@@ -601,6 +634,9 @@ function adjust_opens(cnc,oldcode,code) {
 	for(var li in rgl_modules)
 	    if(cnc.opens[oi]==rgl_modules[li]+oldcode)
 		cnc.opens[oi]=rgl_modules[li]+code;
+	for(var li in wordnet_modules)
+	    if(cnc.opens[oi]==wordnet_modules[li]+oldcode)
+		cnc.opens[oi]=wordnet_modules[li]+code;
 }
 
 function add_concrete2(ix,code) {
@@ -799,7 +835,7 @@ function draw_abstract(g) {
 }
 
 function draw_comment(g) {
-    return div_class("comment",editable("span",text(g.comment || ""),g,edit_comment,"Edit grammar description"));
+    return div_class("comment",editable("span",text(g.comment || "…"),g,edit_comment,"Edit grammar description"));
 }
 
 function module_name(g,ix) {
@@ -864,9 +900,12 @@ function text_mode(g,file,ix) {
 	var ta=node("textarea",{class:"text_mode",rows:25,cols:80},
 		    [text(show_module(g,ix))])
 	var timeout;
-	ta.onkeyup=function() {
+	ta.onkeyup=function(event) {
 	    if(timeout) clearTimeout(timeout);
-	    timeout=setTimeout(function(){parse(ta.value)},400)
+	    timeout=setTimeout(function(){parse(ta.value)},400);
+	    if (event.key == ' ' && event.ctrlKey) {
+	      wordnet_search(g,ta);
+	    }
 	}
 	var mode_button=div_class("right",[button("Guided mode",switch_to_guided_mode)])
 	clear(file)
@@ -1099,13 +1138,15 @@ function draw_concrete(g,i) {
 
 var rgl_modules=["Syntax","Lexicon","Paradigms","Extra","Symbolic"];
 var common_modules=["Prelude"];
+var wordnet_modules=["WordNet"];
 var rgl_info = {
     Prelude: "Some facilities usable in all grammars",
     Paradigms: "Lexical categories (A, N, V, ...) and smart paradigms (mkA, mkN, mkV, ...) for turning raw strings into new dictionary entries.",
     Syntax: "Syntactic categories (Utt, Cl, V, NP, CN, AP, ...), structural words (this_Det, few_Det, ...) and functions for building phrases (mkUtt, mkCl, mkCN, mkVP, mkAP, ...)",
     Lexicon: "A multilingual lexicon with ~350 common words.",
     Extra: "Language-specific extra constructions not available via the common API.",
-    Symbolic: "Functions for symbolic expressions (numbers and variables in mathematics)"
+    Symbolic: "Functions for symbolic expressions (numbers and variables in mathematics)",
+    WordNet: "A large lexicon based on WordNet"
 }
 
 function add_open(g,ci) {
@@ -1131,6 +1172,12 @@ function add_open(g,ci) {
 	    var b=common_modules[i];
 	    add_module(b,b)
 	}
+	if (gfwordnet.languages.indexOf("Parse"+conc.langcode) >= 0) {
+		for(var i in wordnet_modules) {
+			var b=wordnet_modules[i];
+			add_module(b,b+conc.langcode)
+		}
+	}
 	if(list.length>0) {
 	    var file=element("file");
 	    clear(file)
@@ -1424,6 +1471,89 @@ function arg_names(type) {
     return map(unique,names);
 }
 
+function wordnet_search(g,input) {
+	var selection = {
+		current: "Parse"+g.concretes[g.current-1].langcode,
+		langs: {},
+		langs_list: []
+	};
+	if (gfwordnet.languages.indexOf(selection.current) < 0) {
+		return;
+	}
+	var start = input.selectionStart;
+	var end   = input.selectionEnd;
+	if (start == end) {
+		while (start > 0 && input.value.charAt(start-1) != ' ') {
+			start--;
+		}
+	}
+	if (start == end)
+		return;
+	var lemma = input.value.substring(start,end);
+
+	var search_popup = element("search_popup");
+	var viewport = input.getBoundingClientRect();
+	search_popup.style.top     = viewport.top+"px";
+	search_popup.style.left    = viewport.left+"px";
+	search_popup.style.display = "block";
+
+	function close_popup() {
+		search_popup.innerHtml = "";
+		search_popup.style.display = "none";
+		delete(window.onkeyup);
+		delete(window.onclick);
+	}
+	window.onkeyup = function (event) {
+		if (event.keyCode == 27) {
+			close_popup();
+		}
+	}
+	window.onclick = close_popup
+	search_popup.onclick = function (event) {
+		event.stopPropagation();
+	}
+
+	var index=1;
+	for (var i=0; i < g.concretes.length; i++) {
+		var code = g.concretes[i].langcode;
+		var name = "Parse"+code;
+		if (gfwordnet.languages.indexOf(name) >= 0) {
+			selection.langs[name] = {name: langname[code], index: index};
+			selection.langs_list.push(name);
+			index++;
+		}
+	}
+	selection.isEqual = function(other) {
+		if (other.langs_list.length != this.langs_list.length)
+			return false;
+		for (var i = 0; i < other.langs_list.length; i++) {
+			if (other.langs_list[i] != this.langs_list[i])
+				return false;
+		}
+		return true;
+	}
+	gfwordnet.can_select=true;
+	gfwordnet.onclick_select = function(row) {
+		var code   = input.value;
+		var lex_id = row.getAttribute("data-lex-id");
+		input.value= code.substring(0,start)+lex_id+code.substring(end)
+		end = start+lex_id.length;
+		input.selectionStart= end;
+		input.selectionEnd  = end;
+		input.focus();
+
+		var conc=g.concretes[g.current-1];
+		conc.opens || (conc.opens=[]);
+		var open_name = "WordNet"+conc.langcode;
+		if (conc.opens.indexOf(open_name) < 0) {
+			conc.opens.push(open_name);
+		}
+
+		close_popup();
+	}
+	gfwordnet.search(selection, lemma, element("domains"), element("result"), null);
+}
+
 function draw_elin(g,igs,ci,f,dc,df) {
     var fun=f.fun;
     var conc=g.concretes[ci];
@@ -1441,7 +1571,7 @@ function draw_elin(g,igs,ci,f,dc,df) {
 	    }
 	    check_exp(s,check2);
 	}
-	text_editor(el,f.lin,check,true)
+	text_editor(el,f.lin,check,true,wordnet_search.bind(null,g))
     }
     function dl(cls) {
 	var fn=ident(f.fun)
@@ -1773,11 +1903,11 @@ function sort_list(list,olditems,key) {
     }
 }
 
-function text_editor(el,init,ok,async) {
-    string_editor(el,init,ok,async,true)
+function text_editor(el,init,ok,async,search) {
+    string_editor(el,init,ok,async,true,search)
 }
 
-function string_editor(el,init,ok,async,multiline) {
+function string_editor(el,init,ok,async,multiline,search) {
     var p=el.parentNode;
     function restore() {
 	e.parentNode.removeChild(e);
@@ -1808,6 +1938,13 @@ function string_editor(el,init,ok,async,multiline) {
 	var i=node("input",{"class":"string_edit",name:"it",value:init},[]);
 	if(init.length>10) i.size=init.length+5;
     }
+    if (search != null) {
+		i.onkeyup = function(event) {
+			if (event.key == ' ' && event.ctrlKey) {
+				search(i)
+			}
+		};
+    }
     var e=node("form",{},
 	       [i,
 		text(" "),
@@ -1815,6 +1952,9 @@ function string_editor(el,init,ok,async,multiline) {
 		button("Cancel",restore),
 		text(" "),
 		m])
+	if (search != null) {
+		e.appendChild(button("Search",function(event) { search(i); }));
+	}
     e.onsubmit=done
     start("");
 }
